from sqlalchemy.orm import ColumnProperty, class_mapper, DeclarativeBase


class ModelMixin:
    def to_dict(self):
        """Return a dictionary representation of the model instance.

        This method converts the model instance to a dictionary where the keys
        are the column names and the values are the corresponding attribute values.
        This can be useful for serializing tests_models to JSON or other formats.

        Returns:
            dict: A dictionary representation of the model instance.
        """
        return {c.name: getattr(self, c.name) for c in self.__table__.columns}

    @classmethod
    def from_dict(cls, data):
        """
        Creates a new instance of the model from a dictionary of attributes.

        Args:
            data (dict): A dictionary containing the attributes and values for the new instance.

        Returns:
            An instance of the model with the provided attributes.
        """
        return cls(**data)

    @classmethod
    def attribute_names(cls) -> set:
        return set(
            prop.key
            for prop in class_mapper(cls).iterate_properties
            if isinstance(prop, ColumnProperty)
        )




class DbBaseModel(DeclarativeBase, ModelMixin):
    __abstract__ = True

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.to_dict()}>"

    @classmethod
    def attribute_names(cls) -> set:
        return set(
            prop.key
            for prop in class_mapper(cls).iterate_properties
            if isinstance(prop, ColumnProperty)
        )



import logging
from collections import defaultdict
from collections.abc import Sequence
from typing import Type, Generic

from sqlalchemy import (
    select,
    inspect,
    Select,
    asc,
    desc,
    delete,
    text,
)
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import aliased, contains_eager, RelationshipProperty

from .helpers import (
    update_rows,
    prepare_where_clause,
    instance_to_model,
    prepare_update_data, get_record_count,
)
from .decorators import session_manager, autologger
from ..interface.crud import (
    BaseCrudInterface,
    BaseDbModelType,
    BaseSchemaType, SchemaType,
)


from ..interface.session import SessionProviderInterface
from ..models.schema import BaseSchema


logger = logging.getLogger(__name__)

class BaseCrudRepository(Generic[BaseDbModelType, BaseSchemaType], BaseCrudInterface[BaseSchemaType]):
    model: Type[BaseDbModelType] = NotImplemented  # SQLAlchemy модель
    schema: Type[BaseSchemaType] = NotImplemented  # Pydantic схема
    _FIELDS_TO_ALWAYS_EXCLUDE_FROM_CREATE_OR_UPDATE: set = set()

    def __init__(self, provider: SessionProviderInterface) -> None:
        self.session_provider = provider
        self._inspector = inspect(self.model)
        self.pk = {
            key.name: getattr(self.model, key.name)
            for key in self._inspector.primary_key
        }
        self.relations = {}
        self.joins = defaultdict(dict)
        self._add_unique = False
        self.aliases = {}
        self.fields = {
            column.key: getattr(self.model, column.key)
            for column in self._inspector.columns
        }
        self.table = self._inspector.tables[0].name
        for r in self._inspector.relationships.keys():
            # ищем поле с типом relation модели таблицы
            # заполним словарь outerjoin - поля для которых join(Х, isouter=True)
            # заполним словарь fields - поля модели (включая relation) для сортировки
            #
            relation = getattr(self.model, r)
            aliased_relation = aliased(relation.entity.class_)
            if hasattr(relation, "info") and relation.info.get("is_aliased", None):
                self.aliases[f"{relation}"] = aliased_relation
                self.joins[f"{relation}"]["onclause"] = [local_col == getattr(aliased_relation, remote_col.key) for local_col, remote_col in relation.prop.local_remote_pairs]
            self.relations[r] = aliased_relation
            self.joins[f"{relation}"]["isouter"] = not relation.property.innerjoin

            rel_mod_inspect = inspect(relation.entity.class_)
            for column in rel_mod_inspect.columns:
                self.fields[f"{r}.{column.key}"] = getattr(aliased_relation, column.key)

    def _prepare_options_for_model(
        self, schema: Type[BaseSchemaType], options: dict | None = None
    ) -> dict:
        """Формирует словарь с опциями для select операции. Используется
         для загрузки relation:
         ключ словаря - поле для join,
         значение - поля помеченное как contains_eager
         В опциях используем contains_eager, чтобы можно было загрузить
         relation c помощью join
        Args:
            options: словарь полученный на этапе формирования условий where
        Returns:
            dict опций
        """
        if not options:
            options = dict()
        for field in schema.model_fields.keys():
            if not hasattr(self.model, field):
                continue
            column = getattr(self.model, field)
            if column in options:
                continue
            if isinstance(column.property, RelationshipProperty):
                if f"{column}" in self.aliases:
                    options[column] = contains_eager(column, alias=self.aliases[f"{column}"])
                else:
                    options[column] = contains_eager(column)
        return options

    def _prepare_order(
        self,
        order: tuple[str, ...],
    ) -> list:
        """Формирует список сортировки для order_by()
        Args:
            order: список полей для сортировки
             по умолчанию выполняется asc сортировка,
             если имя поля начинается с "-", то desc сортировка
        Returns:
            список колонок сортировки
        """
        order_statements = []

        for field in order:
            direction = asc
            if field.startswith("-"):
                field = field[1:]
                direction = desc
            if "." not in field:
                field = f"{self.table}.{field}"
            order_statements.append(direction(text(field)))
        return order_statements

    def _process_join(self, statement: Select, options: dict) -> Select:
        for join_column in options.keys():
            if f"{join_column}" in self.aliases:
                statement = statement.join(
                    self.aliases[f"{join_column}"],
                    *self.joins[f"{join_column}"]["onclause"],
                    isouter=self.joins[f"{join_column}"]["isouter"]
                )
                self._add_unique = True
                continue
            statement = statement.join(
                join_column, isouter=self.joins[f"{join_column}"]["isouter"]
            )
        statement = statement.options(*list(options.values()))
        return statement

    # @autologger(logger)
    async def _filter_statement_prepare(
        self,
        schema: Type[BaseSchemaType]|Type[SchemaType],
        filters: dict,
        sa_filters: list,
        order: tuple[str, ...],
        limit: int,
        offset: int,
        session: AsyncSession,
        statement: Select | None = None,
        with_count: bool = False,
    ) -> (Select, int):
        """Формирует Select: подготавливает и применяет where, join, options, limit, offset.
        Так же, при необходимости, возвращает общее количество записей для сформированного
        запроса без учета offset, limit.
        Args:
            schema: схема Pydantic для формирования полей необходимых в ответе
            filters: словарь условий из которых формируется where ("field": value, "model.field": value)
            sa_filters: список условий для where, загружаемый непосредственно в where, без обработки
            order: набор полей для сортировки
            session: сессия доступа к БД
            statement: базовое выражение для формирования Select
            with_count: подсчитать общее количество записей для запроса без учета offset, limit
        Returns:
            Select, общее количество записей
        """
        statement = statement or select(self.model)
        where, options = prepare_where_clause(model=self.model, filters=filters, aliases=self.aliases)
        options = self._prepare_options_for_model(schema, options)
        if sa_filters:
            where.extend(sa_filters)
        statement = statement.where(*where)

        count = await get_record_count(session, statement) if with_count else -1
        if options:
            statement = self._process_join(statement, options)
        if offset:
            statement = statement.offset(offset)
        if limit:
            statement = statement.limit(limit)
        if order:
            statement = statement.order_by(*self._prepare_order(order))
        return statement, count

    @session_manager
    @autologger(logger)
    async def create(
        self,
        instance: BaseSchemaType | Sequence[BaseSchemaType],
        *,
        exclude_fields: set | None = None,
        with_relations: bool = False,
        session: AsyncSession = None,
    )->BaseSchemaType| Sequence[BaseSchemaType]:
        """Создает запись в БД из экземпляра схемы
        Args:
            instance: экземпляр схемы данных
            exclude_fields: поля, которые исключаются из model_dump
            with_relations: если передано - то создаем записи в БД для relations
            session: сессия доступа к БД
        Returns:
            экземпляр схемы данных
        """
        db_instances = []
        exclude_fields = exclude_fields or set()
        exclude_fields.update(self._FIELDS_TO_ALWAYS_EXCLUDE_FROM_CREATE_OR_UPDATE)
        if not isinstance(instance, Sequence):
            instances = [
                instance,
            ]
        else:
            instances = instance

        for instance in instances:
            if with_relations:
                for field, value in inspect(self.model).relationships.items():
                    if value.viewonly:
                        continue
                    subinstance = getattr(instance, field)
                    if not subinstance:
                        continue
                    db_instances.append(
                        instance_to_model(
                            instance=subinstance,
                            model=value.mapper.class_,
                            exclude_fields=exclude_fields,
                        )
                    )
            db_instances.append(
                instance_to_model(
                    instance=instance, model=self.model, exclude_fields=exclude_fields
                )
            )
        session.add_all(db_instances)
        # await session.flush()
        return instance

    async def _update(
        self,
        instance: BaseSchemaType | None = None,
        *,
        filters: dict | tuple[str, ...] | None = None,
        exclude_fields: tuple[str, ...] | None = None,
        update_fields: dict | tuple[str, ...] | None = None,
        exclude_none: bool = True,
        session: AsyncSession | None = None,
    ) -> BaseSchemaType:
        """Обновляет запись в БД из экземпляра схемы.
        Args:
            instance: экземпляр схемы данных
            filters: поля (значения берутся из instance) или условия для поиска (словарь, поле:значение), если
              не передано - используются primary key модели
            exclude_fields: поля, которые исключаются из model_dump
            update_fields: если передано tuple - то обновляем только эти поля, если передан dict,
             то используем его для обновления значений
            exclude_none: исключить при дампе схемы Pydantic оля со значением None.
              По умолчанию: True
            session: сессия доступа к БД
        Returns:
            экземпляр схемы данных
        """
        if not instance and not isinstance(update_fields, dict):
            raise ValueError(
                "if instance is none update_fields must be dict for update or vice versa"
            )
        if not filters:
            filters = {pk_f: getattr(instance, pk_f) for pk_f in self.pk.keys()}
        if isinstance(filters, tuple):
            if not instance:
                raise ValueError(f"{instance} is None")
            filters = {f: getattr(instance, f) for f in filters}

        update_data = (
            update_fields
            if isinstance(update_fields, dict)
            else prepare_update_data(
                instance=instance,
                model=self.model,
                exclude_fields=exclude_fields,
                update_fields=update_fields,
                exclude_none=exclude_none,
            )
        )

        results = await self._update_many(
            filters=filters,
            update_data=update_data,
            session=session,
        )
        if len(results) != 1:
            raise ValueError(f"update return not one object: {results}")
        return results[0]

    @session_manager
    @autologger(logger)
    async def update(
        self,
        instance: BaseSchemaType | None = None,
        *,
        filters: dict | tuple[str, ...] | None = None,
        exclude_fields: tuple[str, ...] | None = None,
        update_fields: dict | tuple[str, ...] | None = None,
        session: AsyncSession | None = None,
        exclude_none: bool = True,
    ) -> BaseSchemaType:
        """Обновляет запись в БД из экземпляра схемы.
        Args:
            instance: экземпляр схемы данных. Если в экземпляре присутствуют значения None,
              которые не должны попасть в БД, то нужно передать список полей в параметре
              update_fields
            filters: поля (значения берутся из instance) или условия для поиска (словарь, поле:значение),
              если не передано - используются primary key модели
            exclude_fields: поля, которые исключаются из model_dump
            update_fields: если передано tuple - то обновляем только эти поля, если передан dict,
             то используем его для обновления значений
            exclude_none: исключить при дампе схемы Pydantic поля со значением None.
              По умолчанию: True
            session: сессия доступа к БД
        Returns:
            экземпляр схемы данных
        """
        return await self._update(
            instance,
            filters=filters,
            exclude_fields=exclude_fields,
            update_fields=update_fields,
            exclude_none=exclude_none,
            session=session,
        )

    @session_manager
    @autologger(logger)
    async def update_many(
        self,
        *,
        update_data: dict,
        filters: dict,
        exclude_fields: tuple[str, ...] | None = None,
        session: AsyncSession | None = None,
    ) -> list[BaseSchemaType]:
        """Обновляет несколько строк в БД с использованием переданного словаря.
        Args:
            update_data: словарь полей и значений для обновления, для обновления
             отношений используются вложенные словари
            filters: поля (значения берутся из instance) или условия для поиска
             (словарь, поле:значение), если не передано - используются primary key модели
            exclude_fields: поля, которые исключаются из model_dump
            session: сессия доступа к БД
        Returns:
            экземпляр схемы данных
        """

        return await self._update_many(
            update_data=update_data, filters=filters, session=session
        )

    async def _update_many(
        self,
        *,
        update_data: dict,
        filters: dict,
        session: AsyncSession | None = None,
    ) -> list[BaseSchemaType]:
        """Обновляет несколько строк в БД с использованием переданного словаря.
        Args:
            update_data: словарь полей и значений для обновления, для обновления
             отношений используются вложенные словари
            filters: поля (значения берутся из instance) или условия для поиска
             (словарь, поле:значение), если не передано - используются primary key модели
            session: сессия доступа к БД
        Returns:
            экземпляр схемы данных
        """
        lookup_clause, options = prepare_where_clause(model=self.model, filters=filters, aliases=self.aliases)
        statement = select(self.model)

        options = self._prepare_options_for_model(schema=self.schema, options=options)
        statement = statement.where(*lookup_clause)
        if options:
            statement = self._process_join(statement, options)
        rows = (await session.execute(statement.with_for_update(of=self.model))).scalars()
        if self._add_unique:
            rows = rows.unique()

        results = update_rows(self.schema, rows.all(), update_data)
        return results

    @session_manager
    @autologger(logger)
    async def update_or_create(
        self,
        instance: BaseSchemaType,
        *,
        with_relations: bool = False,
        filters: dict | tuple[str, ...] | None = None,
        exclude_fields: tuple[str, ...] | None = None,
        update_fields: tuple[str, ...] | None = None,
        exclude_none: bool = True,
        session: AsyncSession | None = None,
    ) -> tuple[bool, BaseSchemaType]:
        """Обновляет или создает запись.
        Args:
            instance: экземпляр схемы данных
            with_relations: создать записи в БД для relation из instance
            filters: поля (значения берутся из instance) или условия для поиска (словарь, поле:значение), если
              не передано - используются primary key модели
            exclude_fields: поля, которые исключаются из model_dump
            update_fields: если передано - то обновляем только эти поля
            exclude_none: исключить при дампе схемы Pydantic поля со значением None.
              По умолчанию: True
            session: сессия доступа к БД
        Returns:
            is_created, экземпляр схемы данных
        """

        try:
            instance = await self._update(
                instance,
                filters=filters,
                exclude_fields=exclude_fields,
                update_fields=update_fields,
                exclude_none=exclude_none,
                session=session,
            )
            await session.flush()
            return False, instance
        except ValueError:
            await self.create(instance, with_relations=with_relations, session=session)
            return True, instance

    @session_manager
    @autologger(logger)
    async def delete(
        self, filters: dict, *, session: AsyncSession | None = None
    ) -> int:
        where, _ = prepare_where_clause(model=self.model, filters=filters)
        statement = delete(self.model).where(*where)
        result = await session.execute(statement)
        return result.rowcount

    @session_manager
    async def get_or_none(
        self,
        filters: dict,
        *,
        response_schema: Type[SchemaType] = None,
        order: tuple[str, ...] = None,
        session: AsyncSession | None = None,
    ) -> BaseSchemaType |SchemaType| None:
        result = await self.filter(
            filters=filters,
            session=session,
            response_schema=response_schema,
            order=order,
        )
        return result[0] if len(result) else None

    @session_manager
    async def get(
        self,
        filters: dict,
        *,
        response_schema: Type[SchemaType] = None,
        order: tuple[str, ...] = None,
        session: AsyncSession = None,
    ) -> BaseSchema|SchemaType:
        item = await self.get_or_none(
            response_schema=response_schema,
            filters=filters,
            session=session,
            order=order,
        )
        if not item:
            raise ValueError(f"Not found {self.model} for {filters}")
        return item

    @session_manager
    async def all(
        self,
        response_schema: Type[SchemaType] = None,
        order: tuple[str, ...] = None,
        session: AsyncSession | None = None,
    ) -> list[BaseSchemaType|SchemaType]:
        return await self.filter(
            filters={}, session=session, response_schema=response_schema, order=order
        )

    @session_manager
    async def filter_statement(
        self,
        statement: Select,
        *,
        response_schema: Type[SchemaType] = None,
        session: AsyncSession | None = None,
        is_use_scalars: bool = True,
    ) -> list[BaseSchemaType|SchemaType]:
        """Выбор из БД записей по фильтрам
        Args:
            statement: выражение select
            response_schema: схема в которую будут преобразованы данные из БД
            session: сессия доступа к БД
            is_use_scalars: выставляется в False при использовании в statement
             произвольных возвращаемых полей, а не SqlAlchemy модели
        Returns:
            список экземпляр схемы данных
        """
        schema = response_schema or self.schema
        query = await session.execute(statement)
        if is_use_scalars:
            query = query.scalars()
        if self._add_unique:
            query = query.unique()
        rows = query.all()
        return [schema.model_validate(row) for row in rows]

    @session_manager
    async def filter_statement_with_count(
        self,
        statement: Select,
        *,
        response_schema: Type[SchemaType] = None,
        limit: int | None = None,
        offset: int | None = None,
        session: AsyncSession | None = None,
        count_stmt: Select | None = None,
        is_use_scalars: bool = True,
    ) -> tuple[int, list[BaseSchemaType|SchemaType]]:
        """Выбор из БД записей по фильтрам
        Args:
            statement: выражение select
            response_schema: схема в которую будут преобразованы данные из БД
            limit: max количество возвращаемых записей
            offset: сдвиг от начала в БД
            session: сессия доступа к БД
            count_stmt: отдельный запрос для подсчета общего количества строк,
             может потребоваться в случае сложного запроса с group_by, order_by
            is_use_scalars: выставляется в False при использовании в statement
             произвольных возвращаемых полей, а не SqlAlchemy модели
        Returns:
            список экземпляр схемы данных
        """
        schema = response_schema or self.schema
        #
        # Для отдельных запросов которые используют group_by, order_by
        # необходимо правильно сформировать название полей, сделать это
        # динамически не представляется возможным, поэтому в таких случаях
        # передаем отдельное выражение для общего подсчета строк
        #
        #
        if count_stmt is None:
            count_stmt = statement.with_only_columns(*list(self.pk.values()))
            count_stmt = count_stmt.subquery()
        count = await get_record_count(session, count_stmt)

        if limit:
            statement = statement.limit(limit)
        if offset:
            statement = statement.offset(offset)

        query = await session.execute(statement)
        if is_use_scalars:
            query = query.scalars()
        rows = query.all()
        return count, [schema.model_validate(row) for row in rows]

    # @session_manager
    async def _filter(
        self,
        filters: dict,
        *,
        order: tuple[str, ...] = None,
        sa_filters: list = None,
        response_schema: Type[SchemaType] = None,
        session: AsyncSession | None = None,
        limit: int = 0,
        offset: int = 0,
        statement: Select | None = None,
    ) -> list[BaseSchemaType|SchemaType]:
        """Выбор из БД записей по фильтрам
        Args:
            filters: условия для поиска (словарь, поле:значение)
            order: список полей для упорядочивания
            sa_filters: список условий для where в формате sqlalchemy (не требует обработки)
            response_schema: схема в которую будут преобразованы данные из БД
            session: сессия доступа к БД
            limit: max количество возвращаемых записей
            offset: сдвиг от начала в БД
        Returns:
            список экземпляр схемы данных
        """
        schema = response_schema or self.schema
        statement, _ = await self._filter_statement_prepare(
            schema=schema,
            filters=filters,
            sa_filters=sa_filters,
            order=order,
            limit=limit,
            offset=offset,
            session=session,
            statement=statement,
        )

        return await self.filter_statement(
            statement, response_schema=schema, session=session
        )

    @session_manager
    @autologger(logger)
    async def filter(
        self,
        filters: dict,
        *,
        order: tuple[str, ...] = None,
        response_schema: Type[SchemaType] = None,
        session: AsyncSession | None = None,
        limit: int = 0,
        offset: int = 0,
        sa_filters: tuple = None,
    ) -> list[BaseSchemaType|SchemaType]:
        """Выбор из БД записей по фильтрам
        Args:
            filters: условия для поиска (словарь, поле:значение)
            order: список полей для упорядочивания
            response_schema: схема в которую будут преобразованы данные из БД
            session: сессия доступа к БД
            limit: max количество возвращаемых записей
            offset: сдвиг
            sa_filters: список условий для where в формате sqlalchemy (не требует обработки)
        Returns:
            список экземпляров схемы данных
        """
        return await self._filter(
            filters,
            order=order,
            response_schema=response_schema,
            session=session,
            limit=limit,
            offset=offset,
            sa_filters=sa_filters,
        )

    # @session_manager
    async def _filter_with_total_count(
        self,
        filters: dict,
        *,
        order: tuple[str, ...] = None,
        sa_filters: list = None,
        response_schema: Type[SchemaType] = None,
        session: AsyncSession | None = None,
        limit: int = 0,
        offset: int = 0,
        statement: Select | None = None,
    ) -> (int, list[BaseSchemaType|SchemaType]):
        """Выбор из БД записей по фильтрам, при этом будет дополнительно возвращено
         общее количество записей, соответствующее условиям поиска, без учета limit & offset
        Args:
            filters: условия для поиска (словарь, поле:значение)
            order: список полей для упорядочивания
            sa_filters: список условий для where в формате sqlalchemy (не требует обработки)
            response_schema: схема в которую будут преобразованы данные из БД
            session: сессия доступа к БД
            limit: max количество возвращаемых записей
            offset: сдвиг
        Returns:
            список экземпляров схемы данных
        """
        schema = response_schema or self.schema

        statement, count = await self._filter_statement_prepare(
            schema=schema,
            filters=filters,
            sa_filters=sa_filters,
            order=order,
            limit=limit,
            offset=offset,
            session=session,
            with_count=True,
            statement=statement,
        )

        return count, await self.filter_statement(
            statement, response_schema=schema, session=session
        )

    @session_manager
    async def filter_with_total_count(
        self,
        filters: dict,
        *,
        order: Sequence[str, ...] = None,
        response_schema: Type[SchemaType]= None,
        session: AsyncSession | None = None,
        limit: int = 0,
        offset: int = 0,
        sa_filters: Sequence = None,
    ) -> (int, list[BaseSchemaType|SchemaType]):
        """Выбор из БД записей по фильтрам, при этом будет дополнительно возвращено
         общее количество записей, соответствующее условиям поиска, без учета limit & offset
        Args:
            filters: условия для поиска (словарь, поле:значение)
            order: список полей для упорядочивания
            response_schema: схема в которую будут преобразованы данные из БД
            session: сессия доступа к БД
            limit: max количество возвращаемых записей
            offset: сдвиг
            sa_filters: список условий для where в формате sqlalchemy (не требует обработки)
        Returns:
            список экземпляр схемы данных
        """

        return await self._filter_with_total_count(
            filters,
            order=order,
            response_schema=response_schema,
            session=session,
            limit=limit,
            offset=offset,
            sa_filters=sa_filters,
        )



from collections import defaultdict
from datetime import datetime
from typing import Iterable, Any

from asyncache import cached
from cachetools import TTLCache
from cachetools.keys import hashkey
from sqlalchemy import (
    Column,
    BinaryExpression,
    ColumnOperators,
    Sequence,
    Row,
    RowMapping, func, Select, select,
)
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import contains_eager

from ..interface.crud import BaseDbModelType
from ..interface.crud import BaseSchemaType, BaseSchema


def remove_self(*argv, **kwarg):
    # убираем первый параметр - сессия
    return hashkey(*argv[1:], **kwarg)

@cached(TTLCache(maxsize=1024, ttl=300), key=remove_self)
async def get_record_count(session: AsyncSession, stmt: Select) -> int:
    stmt = select(func.count()).select_from(stmt)
    return (await session.scalars(stmt)).first()

def clause_for_column(field: Column, val) -> BinaryExpression | ColumnOperators:
    """Формирует условие для where()
    Если передано единичное значение или строка, то Column == val
    Если итерируемый объект, то Column.in_(val)
    Args:
        field: колонка
        val: значение
    Returns:
        BinaryExpression | ColumnOperators
    """
    if isinstance(val, Iterable) and not isinstance(val, str):
        vals = list(val)
        if (
            len(vals) == 2
            and isinstance(vals[0], datetime)
            and isinstance(vals[1], datetime)
        ):
            return field.between(vals[0], vals[1])
        return field.in_(val)
    elif isinstance(val, str) and "%" in val:
        return field.ilike(val)
    else:
        return field == val


def process_fields_to_model_and_relation(fields: tuple[str, ...]) -> (set, dict):
    """Преобразует набор полей в:
      - множество (set) полей непосредственно модели ("field")
      - словарь отношений [модель][поле] ("model.field")
    Args:
        fields: набор полей в формате "field" или "model.field"
    Returns:
        (множество полей модели, словарь отношений)
    """
    model = set()
    relation = defaultdict(dict)
    for f in fields:
        if "." in f:
            model_, field = f.split(".", 1)
            relation[model_][field] = 1
        else:
            model.add(f)
    return model, relation


def instance_to_model(
    instance: BaseSchema,
    model: BaseDbModelType,
    exclude_fields: set | None = None,
    update_fields: set | None = None,
) -> BaseDbModelType:
    """Преобразует Pydantic экземпляр в модель SqlAlchemy
    Args:
        instance: Pydantic экземпляр
        model: модель таблицы SqlAlchemy
        exclude_fields: поля которые исключаются из model_dump
        update_fields: если передано - то только эти поля будут выгружены  в модель SqlAlchemy

    Returns:
        экземпляр модели SqlAlchemy
    """
    if not exclude_fields:
        exclude_fields = {}
    # возьмем только те поля которые есть в модели SqlAlchemy
    include_fields: set = model.attribute_names()
    # если заданы update_fields (обновить только эти поля) - берем пересечение с полями модели
    if update_fields:
        include_fields.intersection(update_fields)
    instance = model(
        **instance.model_dump(exclude=exclude_fields, include=include_fields)
    )
    return instance


def prepare_update_data(
    instance: BaseSchema,
    model: BaseDbModelType,
    exclude_fields: tuple[str, ...] | None = None,
    update_fields: tuple[str, ...] | None = None,
    exclude_none: bool = True,
) -> dict:
    """Преобразует Pydantic экземпляр в словарь для метода
     update() используемый для обновления данных в БД
    Args:
        instance: Pydantic экземпляр
        model: модель таблицы SqlAlchemy
        exclude_fields: поля которые исключаются из model_dump()
        update_fields: если передано - то только эти поля будут
         выгружены  в модель SqlAlchemy
        exclude_none: исключить при дампе схемы
         Pydantic поля со значением None. По умолчанию: True
    Returns:
        dict
    """

    update_data = instance.model_dump(exclude_none=exclude_none)
    for field, value in list(update_data.items()):
        if not hasattr(model, field):
            del update_data[field]
        if not isinstance(value, dict):
            continue
        submodel_ = getattr(model, field).entity.class_
        for subfield in list(update_data[field].keys()):
            if hasattr(submodel_, subfield):
                continue
            del update_data[field][subfield]

    result = defaultdict(dict)
    if update_fields:
        for f in update_fields:
            if "." in f:
                model_, field = f.split(".", 1)
                result[model_][field] = update_data[model_][field]
            else:
                result[f] = update_data[f]
        return result
    if exclude_fields:
        for f in exclude_fields:
            if "." in f:
                model_, field = f.split(".", 1)
                if model_ in update_data and field in update_data[model_]:
                    del update_data[model_][field]
            else:
                if f in update_data:
                    del update_data[f]

    return update_data


def prepare_where_clause(
    model: BaseDbModelType,
    filters: dict,
    aliases: dict | None = None,
) -> (list, dict):
    """Формирует список условий поиска для where()
    Ищет поля, в том числе в relations модели SqlAlchemy
    Args:
        model: SqlAlchemy модель таблицы
        filters: список полей и значений для условий (словарь) или список полей
        aliases: словарь c alias для колонок relation
    Returns:
        список условий поиска
    """
    clause = []
    relationships = defaultdict(dict)
    options = dict()

    # Parse filters for direct attributes and relationships
    for attr, value in filters.items():
        if "." in attr:
            # It"s a relationship field
            rel_attr, rel_field = attr.split(".", 1)
            relationships[rel_attr][rel_field] = value
        else:
            # It"s a direct attribute of the model
            column = getattr(model, attr, None)
            clause.append(clause_for_column(column, value))

    # Для загрузки relation будем использовать
    # join совместно contains_eager
    for rel_attr, rel_filter in relationships.items():
        relation = getattr(model, rel_attr)
        if aliases and f"{relation}" in aliases:
            options[relation] = contains_eager(relation, alias=aliases[f"{relation}"])
        else:
            options[relation] = contains_eager(relation)

        for attr, value in rel_filter.items():
            column = getattr(relation.property.mapper.class_, attr)
            clause.append(relation.has(clause_for_column(column, value)))

    return clause, options


def update_row(
    schema: BaseSchemaType, row: Row | RowMapping | Any, data: dict
) -> BaseSchemaType:
    for field, value in data.items():
        # обновим зависимости
        if isinstance(value, dict):
            if not hasattr(row, field):
                raise ValueError(f"Row {row} has no relation {field} from {data=}")
            relation = getattr(row, field)
            for subfield, subvalue in data[field].items():
                setattr(relation, subfield, subvalue)
            continue
        # обновим данные модели
        if not hasattr(row, field):
            raise ValueError(f"Row {row} has no field {field} from {data=}")
        setattr(row, field, value)
    return schema.model_validate(row)


def update_rows(
    schema: BaseSchemaType, rows: Sequence[Row | RowMapping | Any], data: dict
) -> list[BaseSchema]:
    results = []
    for row in rows:
        results.append(update_row(schema, row, data))
    return results



import asyncio
import functools
import logging
from functools import wraps

import sqlalchemy

from ..exceptions.crud import RepositoryIntegrityError
from ..interface.crud import BaseCrudInterface

logger = logging.getLogger(__name__)

def session_manager(func):
    """Декоратор, менеджер сессий - проверяет поле session в kwargs
    и если его нет или значение поля None - открывает сессию, обновляет kwargs
    Args:
        func: имя функции
    Returns:
        декоратор
    """

    @wraps(func)
    async def inner(*args, **kwargs):
        if "session" not in kwargs or kwargs["session"] is None:
            self_: BaseCrudInterface = args[0]
            try:
                async with self_.session_provider.session() as session, session.begin():
                    kwargs["session"] = session
                    result = await func(*args, **kwargs)
            except sqlalchemy.exc.IntegrityError:
                raise RepositoryIntegrityError
        else:
            result = await func(*args, **kwargs)
        return result

    return inner


def autologger(_logger=None, exception_only: bool = True):
    """
    Декоратор для логирования вызовов функции.
    """

    def decorator(func):
        def _log_func_call(bound_logger, func, *args, **kwargs) -> None:
            bound_logger.info(
                f"CALL {func.__name__} {args=} {kwargs=}",
            )

        def _log_func_complete(bound_logger, func, result) -> None:
            bound_logger.info(
                f"COMPLETED {func.__name__} with {result=}",
            )

        def _log_func_error(bound_logger, func, e, *args, **kwargs) -> None:
            bound_logger.exception(
                f"exception {type(e)} - {str(e)}, {args=} {kwargs=}",
            )

        @functools.wraps(func)
        def sync_wrapper(*args, **kwargs):
            bound_logger = _logger or logger
            if not exception_only:
                _log_func_call(bound_logger, func, *args, **kwargs)
            try:
                result = func(*args, **kwargs)
                if not exception_only:
                    _log_func_complete(bound_logger, func, result)
                return result
            except Exception as e:
                _log_func_error(bound_logger, func, e, *args, **kwargs)
                raise

        @functools.wraps(func)
        async def async_wrapper(*args, **kwargs):
            bound_logger = _logger or logger
            if not exception_only:
                _log_func_call(bound_logger, func, *args, **kwargs)
            try:
                result = await func(*args, **kwargs)
                if not exception_only:
                    _log_func_complete(bound_logger, func, result)
                return result
            except Exception as e:
                _log_func_error(bound_logger, func, e, *args, **kwargs)
                raise e

        if asyncio.iscoroutinefunction(func):
            return async_wrapper
        else:
            return sync_wrapper

    return decorator



from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import TypeVar, Generic, Type

from sqlalchemy import Select
from sqlalchemy.ext.asyncio import AsyncSession

from ..interface.session import SessionProviderInterface
from ..models.model import DbBaseModel
from ..models.schema import BaseSchema

BaseDbModelType = TypeVar("BaseDbModelType", bound=DbBaseModel)
BaseSchemaType = TypeVar("BaseSchemaType", bound=BaseSchema)
SchemaType = TypeVar("SchemaType", bound=BaseSchema)

class BaseCrudInterface(Generic[BaseSchemaType], ABC):
    session_provider: SessionProviderInterface

    @abstractmethod
    async def create(
        self,
        instance: BaseSchemaType | Sequence[BaseSchemaType],
        *,
        exclude_fields: set | None = None,
        with_relations: bool = False,
        session: AsyncSession = None,
    )->BaseSchemaType:
        ...

    @abstractmethod
    async def update(
        self,
        instance: BaseSchemaType | None = None,
        *,
        filters: dict | tuple[str, ...] | None = None,
        exclude_fields: tuple[str, ...] | None = None,
        update_fields: dict | tuple[str, ...] | None = None,
        session: AsyncSession | None = None,
    ) -> BaseSchemaType:
        ...

    @abstractmethod
    async def update_or_create(
        self,
        instance: BaseSchemaType,
        *,
        with_relations: bool = False,
        filters: dict | tuple[str, ...] | None = None,
        exclude_fields: tuple[str, ...] | None = None,
        update_fields: tuple[str, ...] | None = None,
        session: AsyncSession | None = None,
    ) -> tuple[bool, BaseSchemaType]:
        ...

    @abstractmethod
    async def update_many(
        self,
        update_data: dict,
        filters: dict,
        exclude_fields: tuple[str, ...] | None = None,
        session: AsyncSession | None = None,
    ) -> list[BaseSchemaType]:
        ...

    @abstractmethod
    async def get_or_none(
        self,
        filters: dict,
        *,
        session: AsyncSession | None = None,
        response_schema: Type[SchemaType] = None,
    ) -> BaseSchemaType | SchemaType | None:
        ...

    @abstractmethod
    async def get(
        self,
        filters: dict,
        response_schema: Type[SchemaType] = None,
        session: AsyncSession = None,
    ) -> BaseSchemaType|SchemaType:
        ...

    @abstractmethod
    async def all(
        self,
        response_schema: Type[SchemaType] = None,
        session: AsyncSession | None = None,
    ) -> list[BaseSchemaType|SchemaType]:
        ...

    @abstractmethod
    async def filter(
        self,
        filters: dict,
        *,
        order: tuple[str, ...] = None,
        response_schema: Type[SchemaType] = None,
        session: AsyncSession | None = None,
        limit: int = 0,
        offset: int = 0,
        sa_filters: tuple = None,
    ) -> list[BaseSchemaType|SchemaType]:
        ...

    @abstractmethod
    async def filter_statement(
        self,
        statement: Select,
        *,
        response_schema: Type[SchemaType] = None,
        session: AsyncSession | None = None,
    ) -> list[BaseSchemaType|SchemaType]:
        ...

    @abstractmethod
    async def filter_statement_with_count(
        self,
        statement: Select,
        *,
        response_schema: Type[SchemaType] = None,
        limit: int | None = None,
        offset: int | None = None,
        session: AsyncSession | None = None,
        count_stmt: Select | None = None,
        is_use_scalars: bool = True,
    ) -> tuple[int, list[BaseSchemaType|SchemaType]]:
        ...

    @abstractmethod
    async def filter_with_total_count(
        self,
        filters: dict,
        *,
        order: Sequence[str, ...] = None,
        response_schema: Type[SchemaType] = None,
        session: AsyncSession | None = None,
        limit: int = 0,
        offset: int = 0,
        sa_filters: Sequence = None,
    ) -> (int, list[BaseSchemaType|SchemaType]):
        ...

    @abstractmethod
    async def delete(
        self, filters: dict, *, session: AsyncSession | None = None
    ) -> int:
        ...



from setuptools import setup, find_packages

setup(
    name='sqlalchemy_pydantic_crud',
    version='0.1.16',
    author='Denisov Pavel',
    author_email='denisov.p.g@gmail.com',
    description='Base CRUD for SqlAlchemy',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/yourusername/your_package_name',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.10',
    install_requires=[
        # List your package dependencies here
    ],
)
